#include <stdio.h>
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#define PIXVAL(iplimagep, x, y) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (x)))
#define PIXVALC(iplimagep, x, y, c) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (x) * 3 + (c)))

//------------------------------------------------------------------ 

enum track_state { TRACK_OFF, TRACK_INIT, TRACK_ON, TRACK_SEL };

struct mouse_param {
    CvBox2D *tboxp;
    enum track_state *tstatep;
};

void on_mouse(int event, int x, int y, int flags, void *param)
{
    static CvPoint topleft;

    struct mouse_param *mpp = (struct mouse_param *)param;

    if (*(mpp->tstatep) == TRACK_SEL) {
        if (x > topleft.x && y > topleft.y) {
            mpp->tboxp->center = cvPoint2D32f((x + topleft.x) / 2,
                                              (y + topleft.y) / 2);
            mpp->tboxp->size = cvSize2D32f(x - topleft.x, y - topleft.y);
        } else {
            mpp->tboxp->center = cvPoint2D32f(topleft.x, topleft.y);
            mpp->tboxp->size = cvSize2D32f(0, 0);
        }
    }

    switch (event) {
    case CV_EVENT_LBUTTONDOWN:
        topleft = cvPoint(x, y);
        mpp->tboxp->center = cvPoint2D32f(x, y);
        mpp->tboxp->size = cvSize2D32f(0, 0);
        mpp->tboxp->angle = 90.0;
        *(mpp->tstatep) = TRACK_SEL;
        break;
    case CV_EVENT_LBUTTONUP:
        if (mpp->tboxp->size.width > 0 && mpp->tboxp->size.height > 0) {
            *(mpp->tstatep) = TRACK_INIT;
        } else {
            *(mpp->tstatep) = TRACK_OFF;
        }
        break;
    }
}

//------------------------------------------------------------------ 
#define R_BINS 16
#define G_BINS 16
#define B_BINS 16
#define BINS (R_BINS * G_BINS * B_BINS)

#define HISTVAL(hist, r, g, b) hist[(r) + ((g) * R_BINS) + ((b) * R_BINS * G_BINS)]

float *create_hist()
{
    return (float *)malloc(sizeof(float) * BINS);
}

void release_hist(float *hist)
{
    free(hist);
}

int calc_hist(IplImage *img, CvBox2D *tboxp, float *hist)
{
    int i, j;
    int ww = (int)(tboxp->size.width / 2);
    int hh = (int)(tboxp->size.height / 2);
    int cx = (int)(tboxp->center.x);
    int cy = (int)(tboxp->center.y);
    int imin = MAX(cx - ww, 0);
    int imax = MIN(cx + ww, img->width - 1);
    int jmin = MAX(cy - hh, 0);
    int jmax = MIN(cy + hh, img->height - 1);
    float sum = 0;
    float w_normalizer = 2 / tboxp->size.width;
    float h_normalizer = 2 / tboxp->size.height;

    memset(hist, 0, sizeof(float) * BINS);

    for (j = jmin; j <= jmax; j++) {
        for (i = imin; i <= imax; i++) {
            float x_norm = (i - cx) * w_normalizer;
            float y_norm = (j - cy) * h_normalizer;
            float r2 = x_norm * x_norm + y_norm * y_norm;
            float weight;
            if (r2 <= 1) { // Epanechnikov profile (up to scales)
                weight = 1 - r2;
            } else {
                weight = 0;
            }

            uchar bidx = PIXVALC(img, i, j, 0) / (256 / B_BINS);
            uchar gidx = PIXVALC(img, i, j, 1) / (256 / G_BINS);
            uchar ridx = PIXVALC(img, i, j, 2) / (256 / R_BINS);
            
            HISTVAL(hist, ridx, gidx, bidx) += weight;
            sum += weight;
        }
    }
    if (sum == 0) {
        return 0;
    }
    for (i = 0; i < BINS; i++) {
        hist[i] /= sum;
    }

    return 1;
}

float bhattacharyya_coeff(float *hist1, float *hist2)
{
    int i;
    float bhatt_coeff = 0;

    for (i = 0; i < BINS; i++) {
        bhatt_coeff += sqrt(hist1[i] * hist2[i]);
    }

    return bhatt_coeff;
}

float likelihood(float *hist1, float *hist2, float lambda)
{
    float bc = bhattacharyya_coeff(hist1, hist2);
    return exp(-lambda * (1 - bc));
}

//------------------------------------------------------------------ 



int main()
{
    int key, i;
    float *hist_model, *hist_cand;

    int nstate = 4;
    int nparticle = 200;
    float lambda = 20.0;
    CvConDensation *cds;
    CvMat *lbound, *ubound;

    hist_model = create_hist();
    hist_cand = create_hist();

    IplImage *frame, *img;
    CvCapture *capture = cvCaptureFromCAM(0);
    if (!capture) {
        fprintf(stderr, "cannot capture\n");
        return 1;
    }

    frame = cvQueryFrame(capture);
    img = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);

    enum track_state tstate;
    CvBox2D tbox;
    struct mouse_param mparam = { &tbox, &tstate };

    cvNamedWindow("track", CV_WINDOW_AUTOSIZE);
    cvSetMouseCallback("track", on_mouse, &mparam);

    cds = cvCreateConDensation(nstate, 0, nparticle);
    lbound = cvCreateMat(nstate, 1, CV_32FC1);
    ubound = cvCreateMat(nstate, 1, CV_32FC1);
    float dynamics[] = { 1, 0, 1, 0,
                         0, 1, 0, 1,
                         0, 0, 1, 0,
                         0, 0, 0, 1 };
    memcpy(cds->DynamMatr, dynamics, nstate * nstate * sizeof(float));

    printf("q: quit\n");
    
    while (1) {
        frame = cvQueryFrame(capture); /* cannot be modified */
        cvFlip(frame, img, NULL);

        if (tstate == TRACK_INIT) {
            calc_hist(img, &tbox, hist_model);
            tstate = TRACK_ON;

            cvmSet(lbound, 0, 0, tbox.center.x - tbox.size.width / 2);
            cvmSet(ubound, 0, 0, tbox.center.x + tbox.size.width / 2);
            cvmSet(lbound, 1, 0, tbox.center.y - tbox.size.height / 2);
            cvmSet(ubound, 1, 0, tbox.center.y + tbox.size.height / 2);
            cvmSet(lbound, 2, 0, -5.0);
            cvmSet(ubound, 2, 0, 5.0);
            cvmSet(lbound, 3, 0, -5.0);
            cvmSet(ubound, 3, 0, 5.0);

            cvConDensInitSampleSet(cds, lbound, ubound);

            cvRandInit(&(cds->RandS[0]), -10, 10, 0);
            cvRandInit(&(cds->RandS[1]), -10, 10, 1);
            cvRandInit(&(cds->RandS[2]), -5, 5, 2);
            cvRandInit(&(cds->RandS[3]), -5, 5, 3);

        } else if (tstate == TRACK_SEL
                   && tbox.size.width > 0 && tbox.size.height > 0) {
            cvEllipseBox(img, tbox, CV_RGB(255, 0, 0), 3);
        } else if (tstate == TRACK_ON) {
            float weight;
            float m0 = 0.0, m1x = 0.0, m1y = 0.0;
            CvBox2D pbox;
            pbox.size = tbox.size;
            pbox.angle = 90.0;
            
            for (i = 0; i < nparticle; i++) {
                int status;
                float x = cds->flSamples[i][0];
                float y = cds->flSamples[i][1];
                pbox.center = cvPoint2D32f(x, y);

                status = calc_hist(img, &pbox, hist_cand);
                if (status == 0) {
                    weight = 0.0;
                } else {
                    weight = likelihood(hist_model, hist_cand, lambda);
                }
                cds->flConfidence[i] = weight;
                m0 += weight;
                m1x += weight * x;
                m1y += weight * y;

                cvCircle(img, cvPoint((int)x, (int)y), 2, CV_RGB(0, 255, 0), -1);
            }
                
            cvConDensUpdateByTime(cds);

            tbox.center = cvPoint2D32f(m1x / m0, m1y / m0);
            cvEllipseBox(img, tbox, CV_RGB(255, 0, 0), 3);
        }
        cvShowImage("track", img);
        if ((key = cvWaitKey(30)) > 0) {
            if (key == 'q' || key == 0x1b) {
                break;
            }
        }
    }

    return 0;
}
